#ifndef _SCSI_IOCTL_H
#define _SCSI_IOCTL_H 

#ifndef _CONFIG_H
#include <linux/config.h>
#endif

#define SCSI_IOCTL_PROBE_HOST 0
#define SCSI_IOCTL_SEND_COMMAND 1

#ifdef CONFIG_BLK_DEV_SD
/* 	Should start at 128	*/
#endif

#ifdef CONFIG_BLK_DEV_SD
/* 	Should start at 256	*/
#endif

#endif


